#include "ourRTSPClient.h"


ourRTSPClient::ourRTSPClient(UsageEnvironment& env, char const* rtspURL, int verbosityLevel, char const* applicationName, portNumBits tunnelOverHTTPPortNum)
  : RTSPClient(env,rtspURL, verbosityLevel, applicationName, tunnelOverHTTPPortNum) 
{
}

ourRTSPClient::~ourRTSPClient() 
{
}

ourRTSPClient* ourRTSPClient::createNew(UsageEnvironment& env, char const* rtspURL,
					int verbosityLevel, char const* applicationName, portNumBits tunnelOverHTTPPortNum) 
{
	return new ourRTSPClient(env, rtspURL, verbosityLevel, applicationName, tunnelOverHTTPPortNum);
}

void continueAfterDESCRIBE(RTSPClient* rtspClient, int resultCode, char* resultString) throw(exception) 
{
	do {
		UsageEnvironment& env = rtspClient->envir(); // alias
		StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs; // alias

		if (resultCode != 0) {
			env << *rtspClient << "Failed to get a SDP description: " << resultString << "\n";
			delete[] resultString;
			exception myException("Falha na aquisio da descrio SDP", 1);
			throw myException;
			break;
		}

		char* const sdpDescription = resultString;
		env << *rtspClient << "Got a SDP description:\n" << sdpDescription << "\n";

		// Cria Sesso
		scs.session = MediaSession::createNew(env, sdpDescription);
		delete[] sdpDescription; // because we don't need it anymore
		if (scs.session == NULL) {
			env << *rtspClient << "Failed to create a MediaSession object from the SDP description: " << env.getResultMsg() << "\n";
			exception myException("Falha em criar MediaSession",1);
			throw myException;
			break;
		} else if (!scs.session->hasSubsessions()) {
			env << *rtspClient << "This session has no media subsessions (i.e., no \"m=\" lines)\n";
			break;
		}

		// Setup das Subsesses. Cada subsesso  uma fonte
		scs.iter = new MediaSubsessionIterator(*scs.session);
		setupNextSubsession(rtspClient);
		return;
	} while (0);

	// Erro
	shutdownStream(rtspClient);
	exception myException("Erro em Descrever Stream",1);
	throw myException;
}

void setupNextSubsession(RTSPClient* rtspClient) throw(exception) 
{
	UsageEnvironment& env = rtspClient->envir(); // alias
	StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs; // alias
  
	scs.subsession = scs.iter->next();
	if (scs.subsession != NULL) {
		if (!scs.subsession->initiate()) {
			env << *rtspClient << "Failed to initiate the \"" << *scs.subsession << "\" subsession: " << env.getResultMsg() << "\n";
			setupNextSubsession(rtspClient); // falhou nessa, vai pra prxima
			char erro[100];
			sprintf(erro,"Falha ao iniciar subsessao %s com resultado %s", *scs.subsession, env.getResultMsg());
			exception myException(erro,1);
			throw myException;
		} else {
			env << *rtspClient << "Initiated the \"" << *scs.subsession
			<< "\" subsession (client ports " << scs.subsession->clientPortNum() << "-" << scs.subsession->clientPortNum()+1 << ")\n";

			// Setupa essa sessao:
			rtspClient->sendSetupCommand(*scs.subsession, continueAfterSETUP);
		}
		return;
	}
  
	//Depois que todas as subsessoes esto configuradas
	//Configura Sink para gravar arquivo
	const char* fileName = scs.getFileToSave();

	scs.qtOut = QuickTimeFileSink::createNew(env, *scs.session, fileName, scs.bufferClient, scs.withClient, scs.heigthClient, scs.frameClient, false,true,false,false);

	if (scs.qtOut == NULL) {
		env << "Failed to create QuickTime file sink for stdout: " << env.getResultMsg();
		scs.qtOut = NULL;
		shutdownStream(rtspClient);
		char erro[100];
		sprintf(erro,"Falha ao criar file sink com resultado: %s", env.getResultMsg());
		exception myException(erro,1);
		throw myException;
	}

	scs.qtOut->startPlaying(subsessionAfterPlaying, scs.subsession);


	// Subsessoes e Sink configurados.
	//RTSP "PLAY" :
	if (scs.session->absStartTime() != NULL) {
		// Special case: The stream is indexed by 'absolute' time, so send an appropriate "PLAY" command:
		rtspClient->sendPlayCommand(*scs.session, continueAfterPLAY, scs.session->absStartTime(), scs.session->absEndTime());
	} else {
		scs.duration = scs.session->playEndTime() - scs.session->playStartTime();
		rtspClient->sendPlayCommand(*scs.session, continueAfterPLAY);
	}
}


void continueAfterSETUP(RTSPClient* rtspClient, int resultCode, char* resultString) throw(exception) 
{
	do {
		UsageEnvironment& env = rtspClient->envir(); // alias
		StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs; // alias

		if (resultCode != 0) {
			env << *rtspClient << "Failed to set up the \"" << *scs.subsession << "\" subsession: " << resultString << "\n";
			char erro[100];
			sprintf(erro,"Falha em configurar a subsessao %s, subsessao: %s", *scs.subsession, resultString);
			exception myException(erro,1);
			throw myException;
			break;
		}

		env << *rtspClient << "Set up the \"" << *scs.subsession
		<< "\" subsession (client ports " << scs.subsession->clientPortNum() << "-" << scs.subsession->clientPortNum()+1 << ")\n";

	} while (0);
	delete[] resultString;

	// Vai para a proxima subsessao:
	setupNextSubsession(rtspClient);
}

void continueAfterPLAY(RTSPClient* rtspClient, int resultCode, char* resultString) throw(exception) 
{
	Boolean success = False;

	do {
		UsageEnvironment& env = rtspClient->envir(); // alias
		StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs; // alias

		if (resultCode != 0) {
			env << *rtspClient << "Failed to start playing session: " << resultString << "\n";
			char erro[100];
			sprintf(erro,"Falha ao comecar a tocar a sessao: %s",resultString);
			exception myException(erro,1);
			throw myException;
			break;
		}

		// Caso queira que a stream tenha um final
		if (scs.duration > 0) {
			unsigned const delaySlop = 2; // number of seconds extra to delay, after the stream's expected duration.  (This is optional.)
			scs.duration += delaySlop;
			unsigned uSecsToDelay = (unsigned)(scs.duration*1000000);
			scs.streamTimerTask = env.taskScheduler().scheduleDelayedTask(uSecsToDelay, (TaskFunc*)streamTimerHandler, rtspClient);
		}

		env << *rtspClient << "Started playing session";
		if (scs.duration > 0) {
			env << " (for up to " << scs.duration << " seconds)";
		}
		env << "...\n";

		success = True;
	} while (0);
	delete[] resultString;

	if (!success) {
		// Erro.
		shutdownStream(rtspClient);
		exception myException("Falha em continueAfterPLAY",1);
		throw myException;
	}
}

// Event handlers:

void subsessionAfterPlaying(void* clientData) {
	MediaSubsession* subsession = (MediaSubsession*)clientData;
	RTSPClient* rtspClient = (RTSPClient*)(subsession->miscPtr);
	StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs; // alias

	// Fecha o SINK das subsessoes:
	if(subsession->sink != NULL) {
		Medium::close(subsession->sink);
		subsession->sink = NULL;
	}

	// Fecha as subsessoes:
	MediaSession& session = subsession->parentSession();
	MediaSubsessionIterator iter(session);
	while ((subsession = iter.next()) != NULL) {
		if (subsession->sink != NULL) return; // this subsession is still active
	}

	// Fecha o Cliente:
	if (!scs.error) {
	shutdownStream(rtspClient);
	}		
}

void subsessionByeHandler(void* clientData) {
	MediaSubsession* subsession = (MediaSubsession*)clientData;
	RTSPClient* rtspClient = (RTSPClient*)subsession->miscPtr;
	UsageEnvironment& env = rtspClient->envir(); // alias

	env << *rtspClient << "Received RTCP \"BYE\" on \"" << *subsession << "\" subsession\n";

	// Agora fecha subsessoes:
	subsessionAfterPlaying(subsession);
}

void streamTimerHandler(void* clientData) {
	ourRTSPClient* rtspClient = (ourRTSPClient*)clientData;
	StreamClientState& scs = rtspClient->scs; // alias

	scs.streamTimerTask = NULL;

	// Fecha o cliente:
	shutdownStream(rtspClient);
}

void shutdownStream(RTSPClient* rtspClient, int exitCode) throw (exception) {
	UsageEnvironment& env = rtspClient->envir(); // alias
	StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs; // alias
	try {
		// Primeira, checa se existe uma subsessao ativa:
		if (scs.session != NULL) {
			if (!scs.error) {
				Boolean someSubsessionsWereActive = False;
				MediaSubsessionIterator iter(*scs.session);
				MediaSubsession* subsession;

				while ((subsession = iter.next()) != NULL) {
					if (subsession->sink != NULL) {
						Medium::close(subsession->sink);
						subsession->sink = NULL;

						if (subsession->rtcpInstance() != NULL) {
							subsession->rtcpInstance()->setByeHandler(NULL, NULL); 
						}

						someSubsessionsWereActive = True;
					}
				}
	
				if (someSubsessionsWereActive) {
					// Manda o comando de TEARDOWN pro servidor fechar a stream.
					rtspClient->sendTeardownCommand(*scs.session, NULL);
				}
			}
		}
		if (!scs.error) {
			if (scs.qtOut != NULL)
			{
				Medium::close(scs.qtOut);
			}
			env << *rtspClient << "Closing the stream.\n";
			Medium::close(rtspClient);
		}
	} catch (exception& myException) {
		throw myException;
	}
}

UsageEnvironment& operator<<(UsageEnvironment& env, const RTSPClient& rtspClient) {
  return env << "[URL:\"" << rtspClient.url() << "\"]: ";
}

UsageEnvironment& operator<<(UsageEnvironment& env, const MediaSubsession& subsession) {
  return env << subsession.mediumName() << "/" << subsession.codecName();
}