#include "ourRTSPConnection.h"
#include <cassert>
#ifdef _DEBUG
#  define _CRTDBG_MAP_ALLOC
#  define _CRTDBG_MAP_ALLOC_NEW
#  include <crtdbg.h>
#  include <assert.h>
#endif

ourRTSPConnection::ourRTSPConnection(void)
{
	ourRtspClient = NULL;
}


ourRTSPConnection::~ourRTSPConnection(void)
{
}


int ourRTSPConnection::initRTSP(const char* url, const char* path, const char* fileName, int withRTSP, int heigthRTSP, int frameRTSP, int bufferRTSP) throw (exception){
	
	//Inicia ambiente

	TaskScheduler* scheduler = BasicTaskScheduler::createNew();
	UsageEnvironment* env = BasicUsageEnvironment::createNew(*scheduler);

	try {
		RTSPClient* rtspClient = openURL(*env, "StartStop", url);
	
		if (rtspClient != NULL) {
			//Adiciona o RTSPClient nas nossas variveis
			ourRtspClient = rtspClient;
			ourURL = url;
			assert(_CrtCheckMemory());	
			//Se RTSPClient foi criado com sucesso, inicia Loop
			StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs;
			scs.filePath = path; 
			scs.fileName = fileName;
			scs.withClient = withRTSP;
			scs.heigthClient = heigthRTSP;
			scs.frameClient = frameRTSP;
			scs.bufferClient = bufferRTSP;
			
			int uSecsToDelay = 60*1000000; // 60s
			scs.packetsTask = env->taskScheduler().scheduleDelayedTask(uSecsToDelay, (TaskFunc*)packetsController, (void*)this);
			env->taskScheduler().doEventLoop(&scs.eventLoopWatchVariable);		
		}
		else {
			exception myException("Problema ao criar o RTSPClient",1);
			throw myException;
		}
		return 0;
	} catch (exception& myExcepiton) {
		if(strcmp(myExcepiton.what(),"Camera desconectada") == 0) {
			ourRtspClient = NULL;
		}
		throw myExcepiton;
	}
	assert(_CrtCheckMemory());	
 }

RTSPClient* ourRTSPConnection::openURL(UsageEnvironment& env, char const* progName, char const* rtspURL) throw (exception){
	try {
		//Procura se j existe na nossa lista
		if(ourRtspClient != NULL) {
			exception myException("J existe uma conexo RTSP aberta",1);
			throw myException;
		}
		RTSPClient* rtspClient = ourRTSPClient::createNew(env, rtspURL, RTSP_CLIENT_VERBOSITY_LEVEL, progName);
		if (rtspClient == NULL) {
			char error[100];
			sprintf(error,"Falha ao criar a RTSPClient para url %s - %s",rtspURL,env.getResultMsg());
			exception myException(error,1);
			throw myException;
		}
		if(!_CrtCheckMemory()) {
			exception myException("Memria Corrompida",1);
			throw myException;
		}
		StreamClientState& scs = ((ourRTSPClient*)rtspClient)->scs;
		rtspClient->sendDescribeCommand(continueAfterDESCRIBE);
		return rtspClient;
	} catch (exception& myException) {
		throw myException;
	}
}

void ourRTSPConnection::packetsController(void* clientData) throw(exception){



	ourRTSPConnection * ourRtspConnection = (ourRTSPConnection*) clientData;



	StreamClientState& scs = ((ourRTSPClient*)ourRtspConnection->ourRtspClient)->scs;


	
	if( ourRtspConnection->ourRtspClient != NULL && scs.session != NULL)
	{
		unsigned newTotNumPacketsReceived = 0;

		MediaSubsessionIterator iter(*scs.session);
		MediaSubsession* subsession;
		while ((subsession = iter.next()) != NULL) {
			RTPSource* src = subsession->rtpSource();
			if (src == NULL) continue;
			newTotNumPacketsReceived += src->receptionStatsDB().totNumPacketsReceived();
		}

		if (newTotNumPacketsReceived == scs.totNumPacketsReceived) {
			// Nenhum pacote foi recebido, faz duas tentativas e desliga stream
			// Se duas tentativas feitas:
			if (scs.timesCheckedPackets == 2) {
				ourRtspConnection->ourRtspClient->envir() << "Closing session, because we stopped receiving packets.\n";
				ourRtspConnection->shutDown();
				ourRtspConnection->ourRtspClient = NULL;
				exception myException("Camera desconectada",0);
				throw myException;
			}
			else {
				scs.timesCheckedPackets++;
				scs.totNumPacketsReceived = newTotNumPacketsReceived;
				// Checa denovo depois de 30s:
				int uSecsToDelay = 10*1000000; // 30s
				scs.packetsTask = ourRtspConnection->ourRtspClient->envir().taskScheduler().scheduleDelayedTask(uSecsToDelay,
                                            (TaskFunc*)packetsController, (void*)ourRtspConnection);
			}
		} else {
			scs.timesCheckedPackets = 0;
			scs.totNumPacketsReceived = newTotNumPacketsReceived;
			// Checa denovo depois de 30s:
			int uSecsToDelay = 30*1000000; // 30s
			scs.packetsTask = ourRtspConnection->ourRtspClient->envir().taskScheduler().scheduleDelayedTask(uSecsToDelay,
                                            (TaskFunc*)packetsController, (void*)ourRtspConnection);	
		}
	}
}

void ourRTSPConnection::shutDown(void) throw (exception) {
	
	try {
		if (ourRtspClient != NULL) {
			StreamClientState& scs = ((ourRTSPClient*)ourRtspClient)->scs;
			scs.eventLoopWatchVariable = 1;
			Sleep(100);
			shutdownStream(ourRtspClient,1);
			ourRtspClient = NULL;
			return;
		}
	} catch (exception& myException) {
		throw myException;
	}
}

void ourRTSPConnection::setPath(const char* path) 
{
	try {
		if(ourRtspClient != NULL) {
			StreamClientState& scs = ((ourRTSPClient*)ourRtspClient)->scs; 
			scs.filePath = path;
		}
	} catch (exception& myException) {
		throw myException;
	}
}